"""MarkdownController implementation for mistune 2.x"""
from __future__ import annotations

import sys
from importlib import import_module
from typing import Any
from typing import Callable
from typing import ClassVar
from typing import Dict
from typing import List
from typing import Optional
from typing import Sequence

import mistune  # type: ignore[import]

from lektor.markdown.controller import MarkdownController
from lektor.markdown.controller import RendererHelper
from lektor.markdown.controller import UnknownPluginError
from lektor.utils import unique_everseen


class ImprovedRenderer(mistune.HTMLRenderer):  # type: ignore[misc]
    lektor: ClassVar = RendererHelper()

    def link(
        self, link: str, text: Optional[str] = None, title: Optional[str] = None
    ) -> str:
        link = self.lektor.resolve_url(link)
        return super().link(link, text, title)

    def image(self, src: str, alt: str = "", title: Optional[str] = None) -> str:
        src = self.lektor.resolve_url(src)
        return super().image(src, alt, title)


if sys.version_info < (3, 8):
    # No typing.TypedDict → punt
    ParserConfigDict = Dict[str, Any]
else:
    from typing import TypedDict

    class ParserConfigDict(TypedDict, total=False):
        block: mistune.BlockParser
        inline: mistune.InlineParser
        plugins: Sequence[Callable[[mistune.Markdown], None]]


MistunePlugin = Callable[[mistune.Markdown], None]


class MarkdownConfig:
    # Enabling these plugins give us feature parity with mistune 0.8.4.
    # We enable them by default for the sake of backwards-compatibility.
    DEFAULT_PLUGINS = ("url", "strikethrough", "footnotes", "table")

    def __init__(self) -> None:
        self.renderer_options = {
            "escape": False,
            "allow_harmful_protocols": True,
        }
        self.renderer_base = ImprovedRenderer
        self.renderer_mixins: List[type] = []
        self.parser_options: ParserConfigDict = {
            "plugins": list(self.DEFAULT_PLUGINS),
        }

    def make_renderer(self) -> ImprovedRenderer:
        bases = tuple(self.renderer_mixins) + (self.renderer_base,)
        renderer_cls = type("renderer_cls", bases, {})
        return renderer_cls(**self.renderer_options)


class MarkdownController2(MarkdownController):
    def make_parser(self) -> mistune.Markdown:
        env = self.env
        cfg = MarkdownConfig()
        # FIXME: call different hooks here for mistune 2?
        env.plugin_controller.emit("markdown-config", config=cfg)
        renderer = cfg.make_renderer()
        env.plugin_controller.emit(
            "markdown-lexer-config", config=cfg, renderer=renderer
        )
        parser_options = cfg.parser_options
        plugins = parser_options.get("plugins")
        if plugins:
            # Resolve and filter out duplicates
            parser_options["plugins"] = tuple(
                unique_everseen(map(self.resolve_plugin, plugins))
            )
        return mistune.Markdown(renderer, **cfg.parser_options)

    PLUGINS = {**mistune.PLUGINS}

    def resolve_plugin(self, plugin: str | MistunePlugin) -> MistunePlugin:
        if callable(plugin):
            return plugin
        if not isinstance(plugin, str):
            raise TypeError(
                f"Plugins should be specified as strings or callables, not {plugin!r}"
            )

        module_name, sep, attr = plugin.partition(":")
        if sep:
            try:
                return getattr(import_module(module_name.strip()), attr.strip())
            except (ImportError, AttributeError) as exc:
                raise UnknownPluginError(f"Can not load plugin {plugin!r}") from exc
        try:
            return self.PLUGINS[plugin]
        except KeyError as exc:
            raise UnknownPluginError(f"Unknown plugin {plugin!r}") from exc



import os
import stat
import sys
import tempfile
from functools import partial
from itertools import chain

__all__ = ["TemporaryDirectory", "importlib_metadata"]


def _ensure_tree_writeable(path: str) -> None:
    """Attempt to ensure that all files in the tree rooted at path are writeable."""
    dirscans = []

    def fix_mode(path, statfunc):
        try:
            # paranoia regarding symlink attacks
            current_mode = statfunc(follow_symlinks=False).st_mode
            if not stat.S_ISLNK(current_mode):
                isdir = stat.S_ISDIR(current_mode)
                fixed_mode = current_mode | (0o700 if isdir else 0o200)
                if current_mode != fixed_mode:
                    os.chmod(path, fixed_mode)
                if isdir:
                    dirscans.append(os.scandir(path))
        except FileNotFoundError:
            pass

    fix_mode(path, partial(os.stat, path))
    for entry in chain.from_iterable(dirscans):
        fix_mode(entry.path, entry.stat)


class FixedTemporaryDirectory(tempfile.TemporaryDirectory):
    """A version of tempfile.TemporaryDirectory that works if dir contains read-only files.

    On python < 3.8 under Windows, if any read-only files are created
    in a TemporaryDirectory, TemporaryDirectory will throw an
    exception when it tries to remove them on cleanup. See
    https://bugs.python.org/issue26660

    This can create issues, e.g., with temporary git repositories since
    git creates read-only files in its object store.

    """

    def cleanup(self) -> None:
        _ensure_tree_writeable(self.name)
        super().cleanup()


if sys.version_info >= (3, 8):
    TemporaryDirectory = tempfile.TemporaryDirectory
    from importlib import metadata as importlib_metadata
else:
    TemporaryDirectory = FixedTemporaryDirectory
    import importlib_metadata



import inspect
import shutil

import pytest

from lektor.project import Project


def write_text(path, text):
    path.parent.mkdir(parents=True, exist_ok=True)
    path.write_text(inspect.cleandoc(text))


def _fixme(*args):
    return pytest.param(*args, marks=pytest.mark.xfail(reason="FIXME"))


@pytest.fixture(scope="module")
def project_path(tmp_path_factory, data_path):
    """Make our own private copy of the demo-project"""
    demo = data_path / "demo-project"
    path = tmp_path_factory.mktemp("test_serve") / "demo-project"

    shutil.copytree(demo, path)

    write_text(path / "assets/TEST.TXT", "Text file.\n")
    return path


@pytest.fixture
def pad(project_path):
    return Project.from_path(project_path).make_env().new_pad()


@pytest.fixture
def asset(pad, asset_path):
    return pad.get_asset(asset_path)


@pytest.mark.parametrize(
    "asset_path, url_path",
    [
        ("/", "/"),
        ("/static", "/static/"),
        ("/static/demo.css", "/static/demo.css"),
        ("/TEST.TXT", "/TEST.txt"),
    ],
)
def test_asset_url_path(asset, url_path):
    assert asset.url_path == url_path


@pytest.mark.parametrize(
    "asset_path, artifact_name",
    [
        ("/", "/"),
        ("/static", "/static"),
        ("/static/demo.css", "/static/demo.css"),
        ("/TEST.TXT", "/TEST.txt"),
    ],
)
def test_asset_artifact_name(asset, artifact_name):
    assert asset.artifact_name == artifact_name


@pytest.mark.parametrize(
    "asset_path, child_names",
    [
        ("/dir_with_index_html", {"index.html"}),
        ("/static", {"demo.css"}),
        ("/static/demo.css", set()),
    ],
)
def test_asset_children(asset, child_names):
    assert set(child.name for child in asset.children) == child_names


@pytest.mark.parametrize("asset_path", ["/static"])
def test_asset_children_no_children_if_dir_unreadable(asset):
    asset.source_filename += "-missing"
    assert len(set(asset.children)) == 0


@pytest.mark.parametrize(
    "asset_path, name, from_url, child_name",
    [
        ("/", "empty", False, "empty"),
        ("/", "missing", False, None),
        ("/", "foo-prefix-makes-me-excluded", False, None),
        (
            "/",
            "_include_me_despite_underscore",
            False,
            "_include_me_despite_underscore",
        ),
        ("/static", "demo.css", False, "demo.css"),
        ("/static/demo.css", "x", False, None),
        ("/empty", "demo.css", False, None),
        # XXX: The from_url special_file_suffixes logic seems not be be used
    ],
)
def test_asset_get_child(asset, name, from_url, child_name):
    if child_name is None:
        assert asset.get_child(name, from_url) is None
    else:
        assert asset.get_child(name, from_url).name == child_name


@pytest.mark.parametrize(
    "asset_path, url_path, expected",
    [
        ("/", ("static",), "/static"),
        ("/", ("static", "demo.css"), "/static/demo.css"),
        ("/", ("missing", "demo.css"), None),
        ("/", ("foo-prefix-makes-me-excluded",), None),
        ("/", ("foo-prefix-makes-me-excluded", "static"), None),
        ("/static", ("demo.css",), "/static/demo.css"),
        _fixme("/", ("TEST.txt",), "/TEST.txt"),
    ],
)
def test_resolve_url_path(asset, url_path, expected):
    if expected is None:
        assert asset.resolve_url_path(url_path) is None
    else:
        assert asset.resolve_url_path(url_path).artifact_name == expected


@pytest.mark.parametrize(
    "asset_path, expected",
    [
        ("/", "<Directory '/'>"),
        ("/static", "<Directory '/static'>"),
        ("/static/demo.css", "<File '/static/demo.css'>"),
        ("/TEST.TXT", "<File '/TEST.txt'>"),
    ],
)
def test_asset_repr(asset, expected):
    assert repr(asset) == expected



import os
from pathlib import Path

import pytest

from lektor.compat import _ensure_tree_writeable
from lektor.compat import FixedTemporaryDirectory
from lektor.compat import TemporaryDirectory


def test_ensure_tree_writeable(tmp_path):
    topdir = tmp_path / "topdir"
    subdir = topdir / "subdir"
    regfile = subdir / "regfile"
    subdir.mkdir(parents=True)
    regfile.touch(mode=0)
    subdir.chmod(0)
    topdir.chmod(0)

    _ensure_tree_writeable(topdir)

    for p in topdir, subdir, regfile:
        assert os.access(p, os.W_OK)


@pytest.mark.parametrize("tmpdir_class", [FixedTemporaryDirectory, TemporaryDirectory])
def test_TemporaryDirectory(tmpdir_class):
    with tmpdir_class() as tmpdir:
        file = Path(tmpdir, "test-file")
        file.touch(mode=0)
        os.chmod(tmpdir, 0)
    assert not os.path.exists(tmpdir)
